import numpy as np
from brainpy.connect import FixedProb
from typing import list

from .res_manager import ResManager
from .base import mpi_size


class Optimizer():

  OPT_GREEDY_INIT = 1

  def __init__(self, opt_method=OPT_GREEDY_INIT, device_memory=40, device_capability=60) -> None:
    self.opt_method = opt_method
    self.device_memory = device_memory
    self.device_capability = device_capability

  def run(self):
    if self.opt_method == self.OPT_GREEDY_INIT:
      self.run_greedy_init()

  def get_edge_weight_matrix(self, syns: list, total_pop_num: int) -> list[[int]]:
    """get edge weight matrix, traffic matrix between populations.

    Parameteresult
    ----------
    syns : list
        synapse list, `ResManager.syns`
    total_pop_num : int
        population total number

    Returns
    -------
    list[[int]]
        two dimensional np array
    """
    edge_weight_matrix = np.zeros((total_pop_num, total_pop_num))
    for syn in syns:
      pre_id = syn.pre.id
      pre_num = syn.pre.shape
      post_id = syn.post.id
      post_num = syn.post.shape

      edge_weight = 0
      conn = syn.conn
      if isinstance(conn, FixedProb):
        edge_weight = pre_num * post_num * conn.prob

      edge_weight = edge_weight_matrix[pre_id][post_id]
      edge_weight_matrix[pre_id][post_id] += edge_weight
    return edge_weight_matrix

  def prepare_input(self) -> tuple(list[[int]], int, int):
    """ prepare optimizer input

    Returns
    -------
    tuple(list[[int]], int, int)
        tuple(edge_weight_matrix, memory_used, memory_capacity)
    """

    pops = ResManager.pops
    total_num = len(pops)

    pop_neuron_count = np.zeros(total_num)
    sum_hardware_store = np.zeros((2, mpi_size))
    sum_hardware_store[0] = sum_hardware_store[0] + self.device_memory
    sum_hardware_store[1] = sum_hardware_store[1] + self.device_capability

    edge_weight_matrix = self.get_edge_weight_matrix(
        ResManager.syns, total_num=total_num)

    memory_used = int((edge_weight_matrix.sum(0) * 10 +
                       pop_neuron_count * 300) / (1024 ** 3))
    memory_capacity = int(sum_hardware_store[0])

    return edge_weight_matrix, memory_used, memory_capacity

  def run_greedy_init(self):
    """Greedy initialization algorithm. Each time find the method that can reduce
    the communication cost between devices after putting it in.
    """

    A, memory_used, memory_capacity = self.prepare_input()
    population_num = A.shape[0]
    assert A.shape == (population_num, population_num)
    assert memory_used.shape == (population_num, )
    device_num = len(memory_capacity)
    assert memory_capacity.shape == (device_num, )
    assert np.all(A >= 0)

    capacity_each = memory_used.sum() / device_num
    memory_capacity[np.where(memory_capacity > capacity_each)] = capacity_each * 1.01
  
    # each population assign to one device
    result = {}
    # memory remaining for each device
    device_available = memory_capacity.copy()

    # move population to decrease traffic cost
    move_gain = np.zeros((device_num, population_num))

    remote_conn = 0
    device_index_pre = 0

    for _ in range(population_num):
      max_index = move_gain.argmax()
      device_index = max_index // population_num
      if device_index_pre != device_index:
        remote_conn = 0
        device_index_pre = device_index
      population_index = max_index % population_num
      if move_gain[device_index, population_index] < 0:
        return result

      r = result.get(device_index, [])
      r.append(population_index)
      result[device_index] = r

      remote_conn_pre = remote_conn
      remote_conn = remote_conn + \
          A[population_index, :].sum() - A[population_index, result.get(device_index)].sum()

      new_capacity = device_available[device_index] - memory_used[population_index] - (
          remote_conn - remote_conn_pre) * 280 / 1024 ** 3

      if new_capacity >= 0:
        device_available[device_index] = new_capacity
        move_gain[device_index] += A[:, population_index] + \
            A[population_index, :]
        move_gain[device_index, memory_used > new_capacity] = -float('inf')
        move_gain[:, population_index] = -float('inf')
      else:
        move_gain[device_index, :] = -float('inf')

        if device_index+1 > device_num-1:
          print("resource not enough")
          break
        device_available[device_index+1] -= memory_used[population_index]
        new_capacity = device_available[device_index+1]
        remote_conn = 0
        move_gain[device_index + 1] += A[:,
                                       population_index] + A[population_index, :]
        move_gain[device_index+1, memory_used > new_capacity] = -float('inf')
        move_gain[:, population_index] = -float('inf')
        r = result.get(device_index + 1, [])
        r.append(population_index)
        result[device_index + 1] = r

      return result

  def save_to_pops_by_rank(self, result: dict):
    d = {}
    for k, v in result.items():
      tmp = []
      for i in v:
        tmp.append(ResManager.pops_by_id(i))
      d[k] = tmp
    return d



from typing import Union, Dict, Callable, Optional

from jax import vmap
from jax.lax import stop_gradient

import brainpy.math as bm
from brainpy.connect import TwoEndConnector, All2All, One2One
from brainpy import dyn
from brainpy.initialize import Initializer, variable_, parameter
from brainpy.integrators import odeint
from brainpy.modes import Mode, BatchingMode, normal
from brainpy.types import Array
from brainpy.dyn.synouts import CUBA
import numpy as np
from bpl.core.base import RemoteDynamicalSystem
from mpi4py import MPI
import jax.numpy as jnp
import mpi4jax


class RemoteExponential(dyn.synapses.Exponential, RemoteDynamicalSystem):
  """Exponential decay synapse model in multi-device environment.
  """

  def __init__(
      self,
      source_rank,
      pre: dyn.NeuGroup,
      target_rank,
      post: dyn.NeuGroup,
      conn: Union[TwoEndConnector, Array, Dict[str, Array]],
      comm=MPI.COMM_WORLD,
      output: dyn.SynOut = CUBA(),
      stp: Optional[dyn.SynSTP] = None,
      comp_method: str = 'sparse',
      g_max: Union[float, Array, Initializer, Callable] = 1.,
      delay_step: Union[int, Array, Initializer, Callable] = None,
      tau: Union[float, Array] = 8.0,
      method: str = 'exp_auto',

      # other parameters
      name: str = None,
      mode: Mode = normal,
      stop_spike_gradient: bool = False,
      
  ):
    super(RemoteExponential, self).__init__(pre=pre,
                                      post=post,
                                      conn=conn,
                                      output=output,
                                      stp=stp,
                                      name=name,
                                      mode=mode,
                                      )
    # parameters
    self.stop_spike_gradient = stop_spike_gradient
    self.comp_method = comp_method
    self.tau = tau
    if bm.size(self.tau) != 1:
      raise ValueError(f'"tau" must be a scalar or a tensor with size of 1. But we got {self.tau}')

    # # connections and weights
    # self.g_max, self.conn_mask = self.init_weights(g_max, comp_method, sparse_data='csr')

    self.comm = comm
    self.source_rank = source_rank
    self.target_rank = target_rank
    self.rank = self.comm.Get_rank()
    if self.rank == source_rank:
      #Make sure the same neuron group only deliver its spike one time during one step network simulation
      if self.pre.name not in self.remote_first_send_mark:
        self.remote_first_send_mark.append(self.pre.name)
        token = mpi4jax.send(self.pre.spike.value, dest=target_rank, tag=0, comm=self.comm)
        self.delay_step = self.remote_register_delay(f"{self.pre.name}.spike", delay_step, self.pre.spike)
    elif self.rank == target_rank:
      # connections and weights
      self.g_max, self.conn_mask = self.init_weights(g_max, comp_method, sparse_data='csr')
      
      if self.pre.name not in self.remote_first_send_mark:
        self.remote_first_send_mark.append(self.pre.name)
        pre_spike, token = mpi4jax.recv(pre.spike, source=source_rank, tag=0, comm=self.comm)
        self.pre_spike = bm.Variable(pre_spike)
        # variables
        self.delay_step = self.remote_register_delay(f"{self.pre.name}.spike", delay_step, self.pre_spike)
      self.g = variable_(bm.zeros, self.post.num, mode)
      self.integral = odeint(lambda g, t: -g / self.tau, method=method)

  def remote_register_delay(
      self,
      identifier: str,
      delay_step: Optional[Union[int, Array, Callable, Initializer]],
      delay_target: bm.Variable,
      initial_delay_data: Union[Initializer, Callable, Array, float, int, bool] = None,
  ):
    """Register delay variable in multi-device enviornmrnt.
    """
    # delay steps
    if delay_step is None:
      delay_type = 'none'
    elif isinstance(delay_step, (int, np.integer, jnp.integer)):
      delay_type = 'homo'
    elif isinstance(delay_step, (bm.ndarray, jnp.ndarray, np.ndarray)):
      if delay_step.size == 1 and delay_step.ndim == 0:
        delay_type = 'homo'
      else:
        delay_type = 'heter'
        delay_step = bm.asarray(delay_step)
    elif callable(delay_step):
      delay_step = parameter(delay_step, delay_target.shape, allow_none=False)
      delay_type = 'heter'
    else:
      raise ValueError(f'Unknown "delay_steps" type {type(delay_step)}, only support '
                       f'integer, array of integers, callable function, brainpy.init.Initializer.')
    if delay_type == 'heter':
      if delay_step.dtype not in [bm.int32, bm.int64]:
        raise ValueError('Only support delay steps of int32, int64. If your '
                         'provide delay time length, please divide the "dt" '
                         'then provide us the number of delay steps.')
      if delay_target.shape[0] != delay_step.shape[0]:
        raise ValueError(f'Shape is mismatched: {delay_target.shape[0]} != {delay_step.shape[0]}')
    if delay_type != 'none':
      max_delay_step = int(bm.max(delay_step))

    # delay target
    if delay_type != 'none':
      if not isinstance(delay_target, bm.Variable):
        raise ValueError(f'"delay_target" must be an instance of Variable, but we got {type(delay_target)}')

    # delay variable
    if delay_type != 'none':
      if identifier not in self.remote_global_delay_data:
        delay = bm.LengthDelay(delay_target, max_delay_step, initial_delay_data)
        self.remote_global_delay_data[identifier] = (delay, delay_target)
        self.local_delay_vars[identifier] = delay
      else:
        delay = self.remote_global_delay_data[identifier][0]
        if delay is None:
          delay = bm.LengthDelay(delay_target, max_delay_step, initial_delay_data)
          self.remote_global_delay_data[identifier] = (delay, delay_target)
          self.local_delay_vars[identifier] = delay
        elif delay.num_delay_step - 1 < max_delay_step:
          self.remote_global_delay_data[identifier][0].reset(delay_target, max_delay_step, initial_delay_data)
    else:
      if identifier not in self.remote_global_delay_data:
        delay = bm.LengthDelay(delay_target, 0)
        self.remote_global_delay_data[identifier] = (delay, delay_target)
        self.local_delay_vars[identifier] = delay
        # Above row of code does not exist in 'register_delay' method. 
        # In multi-device environment, spike of source neuron group need be delivered during local_delay_vars being traversed when Network is updating.
    self.register_implicit_nodes(self.local_delay_vars)
    return delay_step

  def update(self, tdi, pre_spike=None):
    if self.rank == self.target_rank:
      t, dt = tdi['t'], tdi['dt']

      # delays
      if pre_spike is None:
        pre_spike = self.remote_get_delay_data(f"{self.pre.name}.spike", self.delay_step)
      if self.stop_spike_gradient:
        pre_spike = pre_spike.value if isinstance(pre_spike, bm.JaxArray) else pre_spike
        pre_spike = stop_gradient(pre_spike)

      # update sub-components
      self.output.update(tdi)
      if self.stp is not None: self.stp.update(tdi, pre_spike)

      # post values
      if isinstance(self.conn, All2All):
        syn_value = bm.asarray(pre_spike, dtype=bm.dftype())
        if self.stp is not None: syn_value = self.stp(syn_value)
        post_vs = self.syn2post_with_all2all(syn_value, self.g_max)
      elif isinstance(self.conn, One2One):
        syn_value = bm.asarray(pre_spike, dtype=bm.dftype())
        if self.stp is not None: syn_value = self.stp(syn_value)
        post_vs = self.syn2post_with_one2one(syn_value, self.g_max)
      else:
        if self.comp_method == 'sparse':
          f = lambda s: bm.pre2post_event_sum(s, self.conn_mask, self.post.num, self.g_max)
          if isinstance(self.mode, BatchingMode): f = vmap(f)
          post_vs = f(pre_spike)
          # if not isinstance(self.stp, _NullSynSTP):
          #   raise NotImplementedError()
        else:
          syn_value = bm.asarray(pre_spike, dtype=bm.dftype())
          if self.stp is not None: syn_value = self.stp(syn_value)
          post_vs = self.syn2post_with_dense(syn_value, self.g_max, self.conn_mask)
      # updates
      self.g.value = self.integral(self.g.value, t, dt) + post_vs
      # output
      return self.output(self.g)

  def remote_get_delay_data(
      self,
      identifier: str,
      delay_step: Optional[Union[int, bm.JaxArray, jnp.DeviceArray]],
      *indices: Union[int, slice, bm.JaxArray, jnp.DeviceArray],
  ):
    """Get delay data according to the provided delay steps in multi-device enviornment.
    """
    if delay_step is None:
      if bm.ndim(delay_step) == 0:
        return self.remote_global_delay_data[identifier][0](0, *indices)
      else:
        if len(indices) == 0:
          indices = (jnp.arange(delay_step.size),)
        return self.remote_global_delay_data[identifier][0](0, *indices)

    if identifier in self.global_delay_data:
      if bm.ndim(delay_step) == 0:
        return self.remote_global_delay_data[identifier][0](delay_step, *indices)
      else:
        if len(indices) == 0:
          indices = (jnp.arange(delay_step.size),)
        return self.remote_global_delay_data[identifier][0](delay_step, *indices)

    elif identifier in self.local_delay_vars:
      if bm.ndim(delay_step) == 0:
        return self.local_delay_vars[identifier](delay_step)
      else:
        if len(indices) == 0:
          indices = (jnp.arange(delay_step.size),)
        return self.local_delay_vars[identifier](delay_step, *indices)

    else:
      raise ValueError(f'{identifier} is not defined in delay variables.')



import sys
import brainpy.dyn as dyn
from brainpy.modes import Mode, normal
from brainpy.types import Shape, Array
from brainpy.connect import TwoEndConnector
from typing import Union, Sequence, Callable, Tuple, Dict
import jax.tree_util
from brainpy import tools
from .res_manager import ResManager
import bpl

try:
  from mpi4py import MPI
  mpi_size = MPI.COMM_WORLD.Get_size()
  mpi_rank = MPI.COMM_WORLD.Get_rank()
except:
  mpi_size = 1
  mpi_rank = 0


pop_id = 0
def get_pop_id():
  global pop_id
  pop_id += 1
  return pop_id


class BaseNeuron:

  def __init__(
      self,
      shape: Shape,
      *args,
      **kwargs
  ):
    self.args = args
    self.kwargs = kwargs
    self.shape = shape
    # self.model_class = None
    self.lowref = None
    # process id
    self.pid = None
    # population id, autoincrement
    self.id = get_pop_id()
    self.neuron_start_id = 0
    ResManager.pops.append(self)
    ResManager.pops_by_id[self.id] = self

  def __getitem__(self, index: Union[slice, Sequence, Array]):
    return (self, index)

  def __getattr__(self, __name: str):
    return self.lowref.__getattribute__(__name)

  def build(self):  # TODO check current pid
    if self.lowref is not None:
      return self.lowref
    if not hasattr(self, 'model_class'):
      raise Exception("model_class should be assigned")
    self.lowref = self.model_class(self.shape, *self.args, **self.kwargs)
    return self.lowref


class register():
  """decorator class to register user defined neuron model and synapse model to respa and base module

  **Decorator Examples**

    >>> import bpl
    >>> from brainpy.dyn import channels
    >>> @bpl.register()
    >>> class HH(bp.dyn.CondNeuGroup):
    >>>   def __init__(self, size):
    >>>     super(HH, self).__init__(size, )
    >>>     self.INa = channels.INa_TM1991(size, g_max=100., V_sh=-63.)
    >>>     self.IK = channels.IK_TM1991(size, g_max=30., V_sh=-63.)
    >>>     self.IL = channels.IL(size, E=-60., g_max=0.05)
    >>> pop = bpl.HH(100)

  Parameters
  ----------
  name : str = None
    The name of the class in the bpl, respa and base module
    User should use this class to create their model instance.
  model : str = None
    The type of the model
    If model is None the type will be inferred from the class.
  """

  def __init__(self, name: str = None, model: str = None):
    self.name = name
    self.model = model.lower() if model is not None else None

  def __call__(self, cls):
    if self.model is None and issubclass(cls, dyn.NeuGroup) or self.model == 'neuron':
      if self.name is None:
        name = cls.__name__
      else:
        name = self.name
      bases = (BaseNeuron,)
      attrs = {'__qualname__': name, 'model_class': cls}
    elif self.model is None and issubclass(cls, dyn.SynConn) or self.model == 'synapse':
      if self.name is None:
        name = cls.__name__
      else:
        name = self.name
      bases = (BaseSynapse,)
      attrs = {'__qualname__': name,
               'model_class': cls, 'model_class_remote': None}
    else:
      raise RuntimeError(f'custom {self.model} is not supported')
    respa_type = type(name, bases, attrs)
    name_split = __name__.split('.')
    if len(name_split) > 1:
      setattr(sys.modules['.'.join(name_split[:-1])], name, respa_type)
    if len(name_split) > 2:
      setattr(sys.modules['.'.join(name_split[:-2])], name, respa_type)
    setattr(sys.modules[__name__], name, respa_type)
    return cls


class BaseSynapse:

  def __init__(
      self,
      pre: Union[BaseNeuron, Tuple[BaseNeuron, Union[slice, Sequence, Array]]],
      post: Union[BaseNeuron, Tuple[BaseNeuron, Union[slice, Sequence, Array]]],
      conn: Union[TwoEndConnector, Array, Dict[str, Array]],
      *args,
      **kwargs
  ):
    self.pre = pre
    self.post = post
    self.conn = conn
    self.args = args
    self.kwargs = kwargs
    # self.model_class = None
    self.lowref = None
    ResManager.syns.append(self)

  def __getattr__(self, __name: str):
    return self.lowref.__getattribute__(__name)

  def build(self):
    # assert self.pre.lowref is not None and self.post.lowref is not None
    if self.lowref is not None:
      return self.lowref
    if not hasattr(self, 'model_class'):
      raise Exception("model_class should be assigned")

    if isinstance(self.pre, tuple):
      pre_pid = self.pre[0].pid
      pre_shape = self.pre[0].shape
      pre_slice = self.pre[1]
      if pre_pid == mpi_rank:
        pre = self.pre[0].lowref[self.pre[1]]
    elif isinstance(self.pre, BaseNeuron):
      pre_pid = self.pre.pid
      pre_shape = self.pre.shape
      pre_slice = None
      if pre_pid == mpi_rank:
        pre = self.pre.lowref
    else:
      raise ValueError(type(self.pre))

    if isinstance(self.post, tuple):
      post_pid = self.post[0].pid
      post_shape = self.post[0].shape
      post_slice = self.post[1]
      if post_pid == mpi_rank:
        post = self.post[0].lowref[self.post[1]]
    elif isinstance(self.post, BaseNeuron):
      post_pid = self.post.pid
      post_shape = self.post.shape
      post_slice = None
      if post_pid == mpi_rank:
        post = self.post.lowref
    else:
      raise ValueError(type(self.post))

    if pre_pid == post_pid and pre_pid == mpi_rank:
      self.lowref = self.model_class(
          pre, post, self.conn, *self.args, **self.kwargs)
    elif pre_pid == mpi_rank:
      tmp_ = dyn.LIF(0)
      tmp_.size = post_shape
      tmp_.num = tools.size2num(tmp_.size)
      if post_slice is not None:
        tmp_ = tmp_[post_slice]
      self.lowref = self.model_class_remote(
          pre_pid, pre, post_pid, tmp_, conn=self.conn, *self.args, **self.kwargs)
    elif post_pid == mpi_rank:
      tmp_ = dyn.LIF(0)
      tmp_.size = pre_shape
      tmp_.num = tools.size2num(tmp_.size)
      if pre_slice is not None:
        tmp_ = tmp_[pre_slice]
      self.lowref = self.model_class_remote(
          pre_pid, tmp_, post_pid, post, conn=self.conn, *self.args, **self.kwargs)
    return self.lowref


class LIF(BaseNeuron):
  def __init__(
      self,
      shape: Shape,
      *args,
      **kwargs
  ):
    super(LIF, self).__init__(shape, *args, **kwargs)
    self.model_class = dyn.LIF

# another way to define respa LIF
# BaseNeuron.register(dyn.LIF)


class Exponential(BaseSynapse):
  def __init__(
      self,
      pre: Union[BaseNeuron, Tuple[BaseNeuron, Union[slice, Sequence, Array]]],
      post: Union[BaseNeuron, Tuple[BaseNeuron, Union[slice, Sequence, Array]]],
      conn: Union[TwoEndConnector, Array, Dict[str, Array]],
      *args,
      **kwargs
  ):
    super().__init__(pre, post, conn, *args, **kwargs)
    self.model_class = dyn.synapses.Exponential
    self.model_class_remote = None  # dyn.synapses.RemoteExponential


class Network:
  def __init__(self, *ds_tuple, name: str = None, mode: Mode = normal, **ds_dict):
    self.ds_tuple = ds_tuple
    self.ds_dict = ds_dict
    self.lowref = dyn.Network((), name=name, mode=mode)

  def __getattr__(self, __name: str):
    return self.lowref.__getattribute__(__name)

  def build_all_population_synapse(self):
    for pop in ResManager.pops:
      pop.build()
    for syn in ResManager.syns:
      syn.build()
    self.lowref.register_implicit_nodes(
        *map(lambda x: x.lowref, ResManager.pops))
    self.lowref.register_implicit_nodes(
        *map(lambda x: x.lowref, ResManager.syns))

  def build(self):
    self.pops_ = []
    self.syns_ = []

    def reg_pop_syn(v):
      if isinstance(v, BaseNeuron):
        self.pops_.append(v)
      elif isinstance(v, BaseSynapse):
        self.syns_.append(v)
    jax.tree_util.tree_map(reg_pop_syn, self.__dict__)

    def simple_split(pops_):
      res = [[] for i in range(mpi_size)]
      avg = len(pops_) // mpi_size
      for i in range(mpi_size):
        res[i].extend(pops_[i*avg:i*avg+avg])
      res[-1].extend(pops_[avg*mpi_size:])
      return res
    pops_by_rank = simple_split(self.pops_)
    offset = 1
    for i in range(mpi_size):
      for __pops in pops_by_rank[i]:
        __pops.pid = i
    for node in pops_by_rank[mpi_rank]:
      node.neuron_start_id = offset
      offset += node.shape

      node.build()
      self.lowref.register_implicit_nodes(node.lowref)
    for node in self.syns_:
      if node.build() is not None:
        self.lowref.register_implicit_nodes(node.lowref)

  def update(self, *args, **kwargs):
    self.lowref.update(*args, **kwargs)

  def register_nodes(self, *nodes, **named_nodes):
    self.lowref.register_implicit_nodes(*nodes, **named_nodes)

  def register_vars(self, *variables, **named_variables):
    self.lowref.register_implicit_vars(*variables, **named_variables)


class DSRunner:

  def __init__(
      self,
      target: Union[dyn.DynamicalSystem, Network],
      # inputs for target variables
      inputs: Sequence = (),
      fun_inputs: Callable = None,
      # extra info
      dt: float = None,
      t0: Union[float, int] = 0.,
      spike_callback: Callable = None,
      volt_callback: Callable = None,
      **kwargs
  ):
    if not isinstance(target, (Network, dyn.DynamicalSystem)):
      raise ValueError(type(target))

    if isinstance(target, Network):
      target = target.lowref
    
    def _callback(t: float, d: dict):
      for k, v in d.items():
        if k == 'spike' and spike_callback:
          tmp = ''
          for i, j in enumerate(v):
            if j == True:
              tmp += '{},{:.2f}\n'.format(i + 1, t)
          spike_callback(tmp)
        if k == 'V' and volt_callback:
          tmp = ''
          for i, j in enumerate(v):
            tmp += '{},{:.2f},{:.2f}\n'.format(i + 1, t, j)
          volt_callback(tmp)
    c = _callback if spike_callback or volt_callback else None    

    self.lowref = bpl.BplRunner(
      target=target, inputs=inputs,
      fun_inputs=fun_inputs, dt=dt,
      t0=t0, callback=c, **kwargs)

  def __getattr__(self, __name: str):
    return self.lowref.__getattribute__(__name)



import time
from typing import Dict, Union, Sequence, Callable

import numpy as np
import tqdm.auto
from jax.tree_util import tree_map
from jax.experimental.host_callback import id_tap

from brainpy import math as bm
from brainpy import dyn
from brainpy.tools.checking import serialize_kwargs
from brainpy.tools.others.dicts import DotDict


class BplRunner(dyn.DSRunner):
  def __init__(
      self,
      target: dyn.DynamicalSystem,

      # inputs for target variables
      inputs: Sequence = (),
      fun_inputs: Callable = None,

      # extra info
      dt: float = None,
      t0: Union[float, int] = 0.,
      callback: Callable = None,
      **kwargs
  ):
    super(BplRunner, self).__init__(
      target=target, inputs=inputs, fun_inputs=fun_inputs,
      dt=dt, t0=t0, **kwargs)
    self.callback = callback

  def f_predict(self, shared_args: Dict = None):
    if shared_args is None: shared_args = dict()

    shared_kwargs_str = serialize_kwargs(shared_args)
    if shared_kwargs_str not in self._f_predict_compiled:

      monitor_func = self.build_monitors(self._mon_info[0], self._mon_info[1], shared_args)

      def _step_func(t, i, x):
        self.target.clear_input()
        # input step
        shared = DotDict(t=t, i=i, dt=self.dt)
        self._input_step(shared)
        # dynamics update step
        shared.update(shared_args)
        args = (shared,) if x is None else (shared, x)
        out = self.target(*args)
        # monitor step
        mon = monitor_func(shared)
        if self.callback:
          self.callback(t, mon)
        # finally
        if self.progress_bar:
          id_tap(lambda *arg: self._pbar.update(), ())
        return out, mon

      if self.jit['predict']:
        dyn_vars = self.target.vars()
        dyn_vars.update(self.dyn_vars)
        run_func = lambda all_inputs: bm.for_loop(_step_func, dyn_vars.unique(), all_inputs)

      else:
        def run_func(xs):
          # total data
          times, indices, xs = xs

          outputs = []
          monitors = {key: [] for key in (set(self.mon.var_names) | set(self.fun_monitors.keys()))}
          for i in range(times.shape[0]):
            # data at time i
            x = tree_map(lambda x: x[i], xs, is_leaf=lambda x: isinstance(x, bm.JaxArray))

            # step at the i
            output, mon = _step_func(times[i], indices[i], x)

            # append output and monitor
            outputs.append(output)
            for key, value in mon.items():
              monitors[key].append(value)

          # final work
          if outputs[0] is None:
            outputs = None
          else:
            outputs = bm.asarray(outputs)
          for key, value in monitors.items():
            monitors[key] = bm.asarray(value)
          return outputs, monitors
      self._f_predict_compiled[shared_kwargs_str] = run_func
    return self._f_predict_compiled[shared_kwargs_str]



from typing import Union, Dict, List, Tuple
import brainpy.math as bm
from brainpy.base.base import Base


class RemoteDynamicalSystem(Base):
  '''These variable is used in multi-device enviornment.'''
  remote_global_delay_data: Dict[str, Tuple[Union[bm.LengthDelay, None], bm.Variable]] = dict()

  remote_first_send_mark: List[str] = []




class ResManager():
  
  # [BaseNeuron(), ]
  pops = []
  
  # {id: BaseNeuron()}, id is BaseNeuron's id
  pops_by_id = {}
  
  # [BaseSynapse(), ]
  syns = []
  
  # element is an array, like [BaseNeuron(), ], pops_by_rank can be expressed as: {device_index: [BaseNeuron(), ] }
  pops_by_rank = {}



from typing import Union, Dict, Sequence
import brainpy.math as bm
from brainpy import dyn
from brainpy.modes import Mode, normal
from .base import RemoteDynamicalSystem
from mpi4py import MPI
import mpi4jax


class RemoteNetwork(dyn.Network, RemoteDynamicalSystem):
  """Exponential decay synapse model in multi-device environment.
  """

  def __init__(
      self,
      *ds_tuple,
      comm=MPI.COMM_WORLD,
      name: str = None,
      mode: Mode = normal,
      **ds_dict
  ):
    super(RemoteNetwork, self).__init__(*ds_tuple,
                                  name=name,
                                  mode=mode,
                                  **ds_dict)
    self.comm = comm
    if self.comm == None:
      self.rank = None
    else:
      self.rank = self.comm.Get_rank()

  def update_local_delays(self, nodes: Union[Sequence, Dict] = None):
    """overwrite 'update_local_delays' method in Network.
    """
    # update delays
    if nodes is None:
      nodes = tuple(self.nodes(level=1, include_self=False).subset(dyn.DynamicalSystem).unique().values())
    elif isinstance(nodes, dyn.DynamicalSystem):
      nodes = (nodes, )
    elif isinstance(nodes, dict):
      nodes = tuple(nodes.values())
    if not isinstance(nodes, (tuple, list)):
      raise ValueError('Please provide nodes as a list/tuple/dict of dyn.DynamicalSystem.')
    for node in nodes:
      if hasattr(node, 'comm'):
        for name in node.local_delay_vars:
          if self.rank == node.source_rank:
            token = mpi4jax.send(node.pre.spike.value, dest=node.target_rank, tag=3, comm=self.comm)
          elif self.rank == node.target_rank:
            delay = self.remote_global_delay_data[name][0]
            target, token = mpi4jax.recv(node.pre.spike.value, source=node.source_rank, tag=3, comm=self.comm)
            target = bm.Variable(target)
            delay.update(target.value)
      else:
        for name in node.local_delay_vars:
          delay = self.global_delay_data[name][0]
          target = self.global_delay_data[name][1]
          delay.update(target.value)

  def reset_local_delays(self, nodes: Union[Sequence, Dict] = None):
    """overwrite 'reset_local_delays' method in Network.
    """
    # reset delays
    if nodes is None:
      nodes = self.nodes(level=1, include_self=False).subset(dyn.DynamicalSystem).unique().values()
    elif isinstance(nodes, dict):
      nodes = nodes.values()
    for node in nodes:
      if hasattr(node, 'comm'):
        for name in node.local_delay_vars:
          if self.rank == node.source_rank:
            token = mpi4jax.send(node.pre.spike.value, dest=node.target_rank, tag=4, comm=self.comm)
          elif self.rank == node.target_rank:
            delay = self.remote_global_delay_data[name][0]
            target, token = mpi4jax.recv(node.pre.spike.value, source=node.source_rank, tag=4, comm=self.comm)
            target = bm.Variable(target)
            delay.reset(target.value)
      else:
        for name in node.local_delay_vars:
          delay = self.global_delay_data[name][0]
          target = self.global_delay_data[name][1]
          delay.reset(target.value)


# -*- coding: utf-8 -*-

import io
import os
import re
import sys

from setuptools import find_packages
from setuptools import setup

try:
  # require users to uninstall previous brainpy releases.
  import pkg_resources

  installed_packages = pkg_resources.working_set
  for i in installed_packages:
    if i.key == 'brainpy-simulator':
      raise SystemError('Please uninstall the older version of brainpy '
                        f'package "brainpy-simulator={i.version}" '
                        f'(located in {i.location}) first. \n'
                        '>>> pip uninstall brainpy-simulator')
    if i.key == 'brain-py':
      raise SystemError('Please uninstall the older version of brainpy '
                        f'package "brain-py={i.version}" '
                        f'(located in {i.location}) first. \n'
                        '>>> pip uninstall brain-py')
except ModuleNotFoundError:
  pass

# version
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'bpl', '__init__.py'), 'r') as f:
  init_py = f.read()
version = re.search('__version__ = "(.*)"', init_py).groups()[0]

# obtain long description from README
with io.open(os.path.join(here, 'README.md'), 'r', encoding='utf-8') as f:
  README = f.read()

# installation packages
packages = find_packages()
if 'docs' in packages:
  packages.remove('docs')
if 'tests' in packages:
  packages.remove('tests')

# setup
setup(
  name='brainpy-largescale',
  version=version,
  description='brainpy-largescale depends on brainpy',
  long_description=README,
  long_description_content_type="text/markdown",
  author='NanHu Neuromorphic Computing Laboratory Team',
  author_email='nhnao@cnaeit.com',
  packages=packages,
  python_requires='>=3.7',
  install_requires=['numpy>=1.15', 'jax>=0.3.0', 'tqdm', 'brainpy', 'brainpylib', 'numba', 'mpi4py'],
  url='https://github.com/NH-NCL/brainpy-largescale',
  project_urls={
    "Bug Tracker": "https://github.com/NH-NCL/brainpy-largescale/issues",
    "Documentation": "https://brainpy.readthedocs.io/",
    "Source Code": "https://github.com/NH-NCL/brainpy-largescale",
  },
  keywords=('brainpy largescale, '
            'computational neuroscience, '
            'brain-inspired computation, '
            'dynamical systems, '
            'differential equations, ' 
            'brain modeling, '
            'brain dynamics modeling, '
            'brain dynamics programming'),
  classifiers=[
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: Apache Software License',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Software Development :: Libraries',
  ],
  license='Apache-2.0 license',
)
